using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Design;
using System.Windows.Forms;
using HotMon;
using Plux;
using Plux.Framework;
using QAlbum;


[Extension("HotViz", OnReleased = "HotViz_OnReleased")]
[Plug("Startup")]
[ParamValue("ExecuteInMainThread", true, Plug = "Startup")]
[ParamValue("UnplugOnExit", true, Plug = "Startup")]
[Plug("Plux.View", OnPlugged = "View_PlugPlugged")]
[Plug("Plux.Tool", OnPlugged = "Tool_PlugPlugged")]
[Plug("Plux.PropertyGroup")]
[ParamValue("Name", "HotViz", Plugs = new[] { "Plux.View", "Plux.Tool", "Plux.PropertyGroup" })]
[ParamValue("Kind", ViewKind.Tool, Plug = "Plux.View")]
[ParamValue("OrderIndex", 0.9f, Plugs = new[] { "Plux.View", "Plux.Tool", "Plux.PropertyGroup" })]
[ParamValue("Icon", "HotViz.ico", Plugs = new[] {"Plux.View", "Plux.Tool"})]
[Slot("Plux.Preferences", OnPlugged = "Preferences_Plugged", OnUnplugged = "Preferences_Unplugged")]
[Slot("HotMon", OnPlugged = "HotMon_Plugged", OnUnplugged = "HotMon_UnPlugged")]
public class HotViz : ExtensionBase, IStartup, IFlipFormChild, IPropertyGroup, IView, ITool {
  private readonly Display display;
  private readonly Form form;

  public HotViz() {
    display = new Display(GetImage);
    display.Dock = DockStyle.Fill;
    form = new Form();
    InitForm();
  }

  #region ComponentModel Properties

  [Category("HotViz")]
  [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
  //[Editor(@"System.Windows.Forms.Design.StringCollectionEditor, System.Design", typeof(UITypeEditor))]
  //[Editor(@"System.Windows.Forms.Design.StringCollectionEditor, System.Design", typeof(UITypeEditor))]
  [Editor(typeof(MyStringCollectionEditor), typeof(UITypeEditor))]
  public List<String> IgnoreList {
    get { return HotMon != null ? HotMon.IgnoreList : new List<String>(); }
    set {
      if (HotMon != null) {
        HotMon.IgnoreList = value;
      }
    }
  }

  private IHotMon HotMon { get { return Slots == null ? null : Slots["HotMon"].GetPluggedExtension<IHotMon>(); } }

  #endregion

  #region IFlipFormChild Members

  [Browsable(false)]
  public Form Form { get { return form; } }

  [Browsable(false)]
  public IFlipForm FlipFormParent { set { display.flipFormParent = value; } }

  #endregion

  public void Run() {
    Runtime.TaskQueue.InvokeOnEmptied(Show);
  }

  #region ITool Members

  [Browsable(false)]
  public void Dispose() {
    form.Dispose();
  }

  public void Show() {
    if (form.IsDisposed) {
      return;
    }
    form.Show();
    form.Activate();
  }

  #endregion

  #region IView Members

  [Category("HotViz")]
  public string Name { get { return "HotViz"; } }

  [Browsable(false)]
  public Control Control { get { return display; } }

  #endregion

  #region HotViz

  public void HotViz_OnReleased(object sender, ExtensionEventArgs args) {
    display.Hide();
  }

  #endregion

  #region Plug View

  public void View_PlugPlugged(object sender, PlugEventArgs args) {
    if(!Runtime.IsCompositionAllowed) return;
    ToolPlug.Unplug();
    foreach (SlotInfo si in ViewPlug.PluggedInSlots) {
      if (si == args.SlotInfo) {
        continue;
      }
      si.Unplug(ViewPlug);
    }
  }

  #endregion

  #region Plug Tool

  public void Tool_PlugPlugged(object sender, PlugEventArgs args) {
    ViewPlug.Unplug();
    foreach (SlotInfo si in ToolPlug.PluggedInSlots) {
      if (si == args.SlotInfo) {
        continue;
      }
      si.Unplug(ToolPlug);
    }
  }

  #endregion

  private PlugInfo ViewPlug { get { return Plugs["Plux.View"]; } }

  private PlugInfo ToolPlug { get { return Plugs["Plux.Tool"]; } }

  private void OnImageReady(object sender, EventArgs args) {
    if (display == null) {
      return;
    }
    if (display.InvokeRequired) {
      display.BeginInvoke(new EventHandler(OnImageReady), sender, args);
    } else {
      display.Invalidate();
    }
  }

  #region Preferences

  public void Preferences_Plugged(object sender, PlugEventArgs args) {
    var prefs = new Preferences(args);
    if (Plugs["Startup"].PluggedInSlots.Count > 1) {
      prefs.RestoreWindow(form);
    }
  }

  public void Preferences_Unplugged(object sender, PlugEventArgs args) {
    var prefs = new Preferences(args);
    if (Plugs["Startup"].PluggedInSlots.Count > 1) {
      prefs.SaveWindow(form);
    }
  }

  #endregion

  #region InitForm

  private void InitForm() {
    form.Icon = new Icon(Util.GetEmbeddedResourceStream("HotViz.ico"));
    form.KeyPreview = true;
    form.ShowInTaskbar = false;
    form.StartPosition = FormStartPosition.Manual;
    form.Closed += form_Closed;
    form.KeyDown += form_KeyDown;
    form.Controls.Add(display);
  }

  private void form_KeyDown(object sender, KeyEventArgs e) {
    if (IsControlRPressed(e)) {
      display.Invalidate();
    }
  }

  private bool IsControlRPressed(KeyEventArgs e) {
    return e.KeyCode == Keys.R || e.Control;
  }

  private void form_Closed(object sender, EventArgs e) {
    ExitAction.Do(this);
  }

  #endregion

  #region Slot HotMon

  public void HotMon_Plugged(object sender, PlugEventArgs args) {
    IHotMon mon = (IHotMon) args.Extension;
    mon.ImageUpdate += ImageUpdate;
  }

  public void HotMon_UnPlugged(object sender, PlugEventArgs args) {
    IHotMon mon = (IHotMon) args.Extension;
    mon.ImageUpdate -= ImageUpdate;
  }

  private void ImageUpdate(IHotMon sender) {
    OnImageReady(this, EventArgs.Empty);
  }

  private Image GetImage() {
    return HotMon != null ? HotMon.Image : new Bitmap(1, 1);
  }

  #endregion
}

public class MyStringCollectionEditor : System.ComponentModel.Design.CollectionEditor {
  public MyStringCollectionEditor(Type type) : base(type) {
  }

  protected override object CreateInstance(Type itemType) {
    return string.Empty;
  }
}