using System;
using System.Collections.Generic;
using System.ComponentModel;

namespace Plux.Framework {

  [Extension]
  [Plug("Plux.CommandInterpreter")]
  [Slot("Commands", OnPlugged = "CommandsPlugged", OnUnplugged = "CommandsUnplugged", Multiple = true)]
  public class CommandInterpreter : ICommandInterpreter {
    internal readonly Dictionary<string, ICommand> commands = new Dictionary<string, ICommand>();
    private CommandState state;

    private readonly BackgroundWorker executeBackgroundWorker = new BackgroundWorker();
    private readonly BackgroundWorker argumentBackgroundWorker = new BackgroundWorker();

    public CommandInterpreter() {
      commands.Add("help", new HelpCommand(commands));
      commands.Add("execute", new ExecuteCommand(this));
      executeBackgroundWorker.DoWork += ExecuteCommand;
      executeBackgroundWorker.RunWorkerCompleted += ExecuteCommandCompleted;
      argumentBackgroundWorker.DoWork += GetArgument;
      argumentBackgroundWorker.RunWorkerCompleted += GetArgumentCompleted;
    }

    public event EventHandler StateChanged;
    
    public CommandState State {
      get { return state; }
      private set {
        state = value;
        if (StateChanged != null) StateChanged(this, EventArgs.Empty);
      }
    }

    public string Output { get; private set; }

    public bool ExecuteAsync(string cmd) {
      if (state == CommandState.Running) return false;
      State = CommandState.Running;
      string[] cmdArr;
      ICommand c = GetCommand(cmd, out cmdArr);
      if (c == null) {
        Output = cmdArr.Length > 0 ? String.Format("Command '{0}' not found.", cmdArr[0]) : String.Empty;
        State = CommandState.Completed;
        return true;
      }

      executeBackgroundWorker.RunWorkerAsync(new ExecuteCommandInput(c, cmdArr));     
      return true;
    }


    static void ExecuteCommand(object sender, DoWorkEventArgs e) {
      ExecuteCommandInput input = (ExecuteCommandInput)e.Argument;
      e.Result = input.Command.Handler(input.CommandArray);
    }

    void ExecuteCommandCompleted(object sender, RunWorkerCompletedEventArgs e) {
      Output = (string) e.Result;
      State = CommandState.Completed;
    }

    public string Execute(string cmd) {
      string[] cmdArr;
      ICommand c = GetCommand(cmd, out cmdArr);
      if (c == null)
        return cmdArr.Length > 0 ? String.Format("Command '{0}' not found.", cmdArr[0]) : String.Empty;
      return c.Handler.Invoke(cmdArr);
    }

    private ICommand GetCommand(string cmd, out string[] cmdArr) {
      cmdArr = cmd.Split(new char[] { ' ', '|' }, StringSplitOptions.RemoveEmptyEntries);
      return commands.ContainsKey(cmdArr[0]) ? commands[cmdArr[0]] : null;
    }

    public string[] GetCommandOptions(ref string cmdStart) {
      string[] command = cmdStart.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
      int pos;
      if (command.Length == 0)
        pos = 0;
      else if (cmdStart[cmdStart.Length - 1] != ' ') {
        pos = command.Length - 1;
        cmdStart = cmdStart.Substring(0, cmdStart.Length - command[pos].Length);
      }
      else
        pos = command.Length;

      IEnumerable<string> complOption = null;
      if (pos == 0) complOption = commands.Keys;
      else {
        string c = command[0];
        if (commands.ContainsKey(c))
          complOption = commands[c].GetArguments(command, pos - 1);
      }
      if (complOption == null) return new string[0];
      List<string> complList = new List<string>();
      if (command.Length == pos) {
        complList.AddRange(complOption);
      } else {
        string startStr = command[pos];
        int strLen = command[pos].Length;
        foreach (string com in complOption) {
          if (com.Length >= strLen &&
              com.Substring(0, strLen).Equals(
                  startStr, StringComparison.CurrentCultureIgnoreCase))
            complList.Add(com);
        }
      }
      return complList.ToArray();
    }

    void GetArgumentCompleted(object sender, RunWorkerCompletedEventArgs e) {
      throw new NotImplementedException();
    }

    void GetArgument(object sender, DoWorkEventArgs e) {
      throw new NotImplementedException();
    }

    public void CommandsPlugged(object sender, PlugEventArgs args) {
      IConsoleCommands consoleCommands = (IConsoleCommands)args.Extension;
      foreach (ICommand command in consoleCommands.Commands) {
        if (!commands.ContainsKey(command.Name))
          commands.Add(command.Name, command);
        else {
          args.SlotInfo.ExtensionInfo.LogEvent(
            new LogEventArgs(
              MessageType.Warning, VerbosityLevel.Normal,
              "Error on adding command: \"{0}\". Command already exists.",
              command.Name));
        }
      }
    }

    public void CommandsUnplugged(object sender, PlugEventArgs args) {
      IConsoleCommands consoleCommands = (IConsoleCommands) args.Extension;
      foreach (ICommand command in consoleCommands.Commands) {
        if (commands.ContainsValue(command))
          commands.Remove(command.Name);
      }
    }
  }

  class ExecuteCommandInput {
    public ICommand Command { get; set; }
    public string[] CommandArray { get; set; }
    public ExecuteCommandInput(ICommand command, string[] commandArray) {
      Command = command;
      CommandArray = commandArray;
    }
  }
  
}
