﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO.IsolatedStorage;
using System.IO;

namespace Plux.Framework {

  public abstract class CompositionPersistor {
    protected static IsolatedStorageFile store;
    protected const string VERSION1 = "Plux Compilation Version 1.1";
    protected Composer composer = null;

    static CompositionPersistor() {
      store = GetStore();
    }

    protected static IsolatedStorageFile GetStore() {
      return IsolatedStorageFile.GetStore(
        IsolatedStorageScope.User | IsolatedStorageScope.Assembly, null, null);
    }

    public static void DeleteFile(string filename)
    {
      store.DeleteFile(filename);
    }

    public static void OpenFileLocation()
    {
      IsolatedStorageFileStream stream = new IsolatedStorageFileStream("temp", FileMode.Create, store);
      string path = stream.GetType().GetField("m_FullPath", System.Reflection.BindingFlags.Instance | System.Reflection.BindingFlags.NonPublic).GetValue(stream).ToString();
      stream.Close();
      store.DeleteFile("temp");

      System.Diagnostics.Process.Start("explorer", path.Substring(0,path.LastIndexOf('\\') + 1));
    }
  }

  public class CompositionSaver : CompositionPersistor, IDisposable {

    StreamWriter writer;
    IList<SlotInfo> slotsToSave;
    ExtensionInfo extension;
    string description;

    private CompositionSaver(ExtensionInfo ext, IEnumerable<SlotInfo> slots, string description, string filename) {
      extension = ext;
      this.description = description;
      writer = new StreamWriter(new IsolatedStorageFileStream(filename, FileMode.Create, store));
      slotsToSave = new List<SlotInfo>(slots);
    }

    public void Dispose() {
      writer.Dispose();
    }

    public static void Save(ExtensionInfo ext, string description) {
      Save(ext, ext.SlotInfos, description);
    }

    public static void Save(ExtensionInfo ext, IEnumerable<SlotInfo> slots, string description) {
      using (var saver = new CompositionSaver(ext, slots, description, System.IO.Path.GetRandomFileName())) {
        saver.Save();
      }
    }

    private void Save() {
      try {
        writer.WriteLine(VERSION1);
        writer.WriteLine(description);
        if(extension.IsPlugged) {
          writer.Write("ROOT;{0}", extension.IsShared ? "S" : "U");
          WriteExtensionType(extension.ExtensionTypeInfo);
          SaveSlots();
        }
      } finally {
        writer.Close();
      }
    }

    private void WriteExtensionType(ExtensionTypeInfo type) {
      writer.Write(";{0};{1};{2};{3}", 
        type.Name, 
        type.PluginInfo.Name, 
        type.PluginInfo.RepositoryId.Uri,
        type.PluginInfo.RepositoryId.LastWriteTime);
    }

    private void SaveSlots() {
      foreach (SlotInfo s in slotsToSave) {
        if (s.IsOpen) {
          WriteSlotProperties(s);
        }
      }
      while(slotsToSave.Count > 0) {
        SlotInfo s = slotsToSave[0];
        slotsToSave.RemoveAt(0);
        writer.WriteLine();
        SaveRegisteredPlugs(s);
        SavePluggedPlugs(s);
        SaveSelectedPlugs(s);
      }
    }

    private void SaveRegisteredPlugs(SlotInfo slot) {
      foreach (PlugTypeInfo p in slot.RegisteredPlugTypeInfos) {
        writer.Write("EXT R;");
        WriteExtensionType(p.ExtensionTypeInfo);
      }
    }

    private void SavePluggedPlugs(SlotInfo slot) {
      foreach (PlugInfo p in slot.PluggedPlugInfos) {
        writer.Write("EXT P;{0}", p.ExtensionInfo.IsShared ? "S" : "U");
        WriteExtensionType(p.PlugTypeInfo.ExtensionTypeInfo);
        foreach (SlotInfo s in SlotOpenHistory.GetOpenHistory(p.ExtensionInfo)) {
        //foreach (SlotInfo s in p.ExtensionInfo.SlotInfos) {
          if (s.IsOpen) {
            WriteSlotProperties(s);
            slotsToSave.Add(s);
          }
        }
      }
    }

    private void SaveSelectedPlugs(SlotInfo slot) {
      foreach (PlugInfo p in slot.SelectedPlugs) {
        writer.Write("EXT S;");
        WriteExtensionType(p.PlugTypeInfo.ExtensionTypeInfo);
      }
    }

    private void WriteSlotProperties(SlotInfo slot) {
      writer.Write("SLOT AUTOOPEN={0};AUTOPLUG={1};AUTOREGISTER={2};AUTORELEASE={3};LAZYLOAD={4};UNIQUE={5};SELECTIONMODE={6};{7}", 
        slot.AutoOpen ? 't' : 'f',
        slot.AutoPlug ? 't' : 'f',
        slot.AutoRegister ? 't' : 'f',
        slot.AutoRelease ? 't' : 'f',
        slot.LazyLoad ? 't' : 'f',
        slot.Unique ? 't' : 'f',
        slot.SelectionMode,
        slot.Name);
    }
  }

  public class CompositionRestorer : CompositionPersistor, IDisposable {

    StreamReader reader;
    IList<SlotInfo> slotsToRestore;
    private bool ignoreExisting;

    private CompositionRestorer(string filename, bool ignoreExisting) {
      reader = new StreamReader(new IsolatedStorageFileStream(filename, FileMode.Open, store));
      slotsToRestore = new List<SlotInfo>();
      this.ignoreExisting = ignoreExisting;
    }

    public void Dispose() {
      reader.Dispose();
    }

    public static void Restore(String filename, ExtensionInfo root)
    {
        Restore(filename, root, false);
    }

    public static void Restore(String filename, ExtensionInfo root, bool ignoreExisting) {
      using(CompositionRestorer restorer = new CompositionRestorer(filename, ignoreExisting)) {
        restorer.Restore(root);
      }
    }

    private void Restore(ExtensionInfo root) {
      try {
        string version = reader.ReadLine();
        if(version != VERSION1) {
          throw new Exception("Invalid file version");
        }
        string description = reader.ReadLine();
        Runtime.AutoOpen = Runtime.AutoRegister = Runtime.AutoPlug = false;
        composer = Runtime.DisableComposition();
        
        SlotInfo slot = null;
        while(!reader.EndOfStream) {
          RestoreLine(reader.ReadLine(), slot, root);
          if(slotsToRestore.Count > 0) {
            slot = slotsToRestore[0];
            slotsToRestore.RemoveAt(0);
          } else {
            slot = null;
          }
        }
      }
      finally {
        try {
        if(composer != null) composer.EnableComposition();
        } finally {
          if(reader != null) reader.Close();
        }
        Runtime.AutoOpen = Runtime.AutoRegister = Runtime.AutoPlug = true;
      }
    }

    private void RestoreLine(string extensionsAsString, SlotInfo slot, ExtensionInfo root) {
      string[] extensions = extensionsAsString.Split(new String[] { "EXT " }, StringSplitOptions.RemoveEmptyEntries);
      foreach(string extAsString in extensions) {
        RestoreExtension(extAsString, slot, root);
      }
    }

    private void RestoreExtension(string extensionAsString, SlotInfo slot, ExtensionInfo root) {
      var slots = extensionAsString.Split(new String[] { "SLOT " }, StringSplitOptions.RemoveEmptyEntries);
      string[] remainingSlots = new string[slots.Length-1];
      Array.Copy(slots, 1, remainingSlots, 0, remainingSlots.Length);
      var info = ExtensionRestoreInfo.Parse(slots[0]);
      var plugin = GetPluginInfo(info);
      if(plugin == null) {
        throw new Exception(String.Format("Plugin {0} not found", info.PluginName));
      }
      var extensionType = plugin.ExtensionTypeInfos[info.ExtensionTypeName];
      if(info.IsRoot) {
        if(extensionType.ExtensionInfos.Count == 0) {
          throw new Exception(String.Format("Restoration root extension {0} not found in composition", extensionType.Name));
        }
        //RestoreSlots(extensionType.ExtensionInfos[0], remainingSlots); // TODOFIX: was ist wenn es mehrere gibt?
        RestoreSlots(root, remainingSlots);
      } else if(slots[0].StartsWith("R")) {
        composer.Register(slot, extensionType.PlugTypeInfos[slot.Name]);
      } else if(slots[0].StartsWith("P")) {
          ExtensionInfo ext;
          if (info.IsShared)
              ext = composer.GetSharedExtension(extensionType, true);
          else //Extension is unique
          {
              if (extensionType.ExtensionInfos.Count != 0)
                  ext = this.ignoreExisting ? extensionType.ExtensionInfos[0] : composer.CreateUniqueExtension(extensionType);
              else
                  ext = composer.CreateUniqueExtension(extensionType);
          }
          composer.Plug(slot, ext.PlugInfos[slot.Name]);

          if(remainingSlots.Length > 0) {
            RestoreSlots(ext, remainingSlots);
        }
      } else if(slots[0].StartsWith("S")) {
        var newestExtension = extensionType.ExtensionInfos[extensionType.ExtensionInfos.Count - 1];
        composer.SelectPlugs(slot, newestExtension.PlugInfos[slot.Name]);
      }
    }
    
    private void RestoreSlots(ExtensionInfo ext, IEnumerable<string> slotsAsString) {
      foreach(string s in slotsAsString) {
        var info = SlotRestoreInfo.Parse(s);
        RestoreSlot(ext.SlotInfos[info.Name], info);
      }
    }

    private void RestoreSlot(SlotInfo slot, SlotRestoreInfo info) {
      composer.SetAutoOpen(slot, info.AutoOpen);
      composer.SetAutoPlug(slot, info.AutoPlug);
      composer.SetAutoRegister(slot, info.AutoRegister);
      composer.SetAutoRelease(slot, info.AutoRelease);
      composer.SetLazyLoad(slot, info.LazyLoad);
      composer.SetUnique(slot, info.Unique);
      composer.SetSelectionMode(slot, info.selectionMode);
      if (!slot.IsOpen)
        composer.Open(slot);
      slotsToRestore.Add(slot);
    }

    private class ExtensionRestoreInfo {
      public String ExtensionTypeName;
      public String PluginName;
      public Uri Uri;
      public DateTime? LastWriteTime { get; set; }
      public bool IsShared;
      public bool IsRoot;
      private ExtensionRestoreInfo() {}
      public static ExtensionRestoreInfo Parse(string s) {
        var r = new ExtensionRestoreInfo();
        r.IsRoot = s.StartsWith("ROOT");
        string[] properties = s.Substring(s.IndexOf(';')+1).Split(';');
        r.IsShared = (properties[0] == "S");
        r.ExtensionTypeName = properties[1];
        r.PluginName = properties[2];
        r.Uri = new Uri(properties[3]);
        r.LastWriteTime = DateTime.Parse(properties[4]);
        return r;
      }

      private static string FormatDateTime(DateTime? dateTime) {
        return dateTime.ToString();
      }

      private static DateTime? ParseDateTime(String dateTime) {
        int day, month, year, hour, minute, second;
        String[] help = dateTime.Split(' ');
        String[] date = help[0].Split('.');
        String[] time = help[1].Split(':');

        day = Convert.ToInt32(date[0]);
        month = Convert.ToInt32(date[1]);
        year = Convert.ToInt32(date[2]);
        hour = Convert.ToInt32(time[0]);
        minute = Convert.ToInt32(time[1]);
        second = Convert.ToInt32(time[2]);

        return new DateTime?(new DateTime(year, month, day, hour, minute, second));
      }
    }

    private PluginInfo GetPluginInfo(ExtensionRestoreInfo info) {
      PluginInfo result = null;
      foreach(PluginInfo p in Runtime.Repository.PluginInfos.Find(info.PluginName)) {
        if(!p.RepositoryId.Uri.Equals(info.Uri)) continue;
        //if(!p.RepositoryId.LastWriteTime.Equals(info.LastWriteTime)) continue; // stimmt nicht genau ueberein, wegen millisek?
        result = p;
        break;
      }
      return result;
    }

    private class SlotRestoreInfo {
      public bool AutoOpen; 
      public bool AutoPlug;
      public bool AutoRegister;
      public bool AutoRelease;
      public bool LazyLoad;
      public bool Unique;
      public SelectionMode selectionMode;
      public String Name;

      private SlotRestoreInfo() { }

      public static SlotRestoreInfo Parse(string s) {
        String[] properties = s.Split(';');
        var r = new SlotRestoreInfo();
        r.AutoOpen = (properties[0].EndsWith("=t"));
        r.AutoPlug = (properties[1].EndsWith("=t"));
        r.AutoRegister = (properties[2].EndsWith("=t"));
        r.AutoRelease = (properties[3].EndsWith("=t"));
        r.LazyLoad = (properties[4].EndsWith("=t"));
        r.Unique = (properties[5].EndsWith("=t"));
        r.selectionMode = properties[6].EndsWith("Single") ? SelectionMode.Single : SelectionMode.Multiple;
        r.Name = properties[7];
        return r;
      }
    }

    public static List<SavedCompilationInfo> GetSavedCompilations() {
      List<String> storedFiles = new List<String>();
      storedFiles.AddRange(store.GetFileNames());
      List<SavedCompilationInfo> savedComp = new List<SavedCompilationInfo>();
      foreach (string filename in storedFiles)
          savedComp.AddRange(GetRootExtensionsOfFile(filename));
      return savedComp;
    }

    public static List<SavedCompilationInfo> GetSavedCompilationInfo(string filename)
    {
        return GetRootExtensionsOfFile(filename);
    }

    private static List<SavedCompilationInfo> GetRootExtensionsOfFile(string filename)
    {
        List<SavedCompilationInfo> sci = new List<SavedCompilationInfo>();
        StreamReader reader = null;

        reader = new StreamReader(new IsolatedStorageFileStream(filename, FileMode.Open, store));
        reader.ReadLine(); //Version
        string desc = reader.ReadLine(); //Description
        string rootLine = reader.ReadLine();
        string[] splittedRootLine = rootLine.Split(new String[] { "SLOT " }, StringSplitOptions.RemoveEmptyEntries);
        string[] rootExt = splittedRootLine[0].Split(';');
        List<SlotInfo> rootSlots = new List<SlotInfo>();
        PluginInfo plugin = GetPluginInfo(rootExt[3], new Uri(rootExt[4]));
        if (plugin != null)
        {
            var extensionType = plugin.ExtensionTypeInfos[rootExt[2]];
            if (extensionType.ExtensionInfos.Count != 0)
            {
                //--------rootSlots---------------
                for (int i = 1; i < splittedRootLine.Length; ++i)
                {
                    string[] slot = splittedRootLine[i].Split(';');
                    string slotName = slot[slot.Length - 1];
                    rootSlots.Add(extensionType.ExtensionInfos[0].SlotInfos[slotName]);
                }
                for (int i = 0; i < extensionType.ExtensionInfos.Count; ++i)
                    sci.Add(new SavedCompilationInfo(filename, desc, extensionType.ExtensionInfos[i], rootSlots));
            }
            else
                sci.Add(new SavedCompilationInfo(filename, desc, null, null));
        }
        else
            sci.Add(new SavedCompilationInfo(filename, desc, null, null));
        reader.Dispose();
        
        return sci;
    }

    private static PluginInfo GetPluginInfo(string pluginName, Uri uri)
    {
        PluginInfo result = null;
        foreach (PluginInfo p in Runtime.Repository.PluginInfos.Find(pluginName))
        {
            if (!p.RepositoryId.Uri.Equals(uri)) continue;
            //if(!p.RepositoryId.LastWriteTime.Equals(info.LastWriteTime)) continue; // stimmt nicht genau ueberein, wegen millisek?
            result = p;
            break;
        }
        return result;
    }

    public class SavedCompilationInfo : IComparable
    {
        private string storedFile;
        private string description;
        private ExtensionInfo root;
        private List<SlotInfo> rootSlots;

        public string StoredFile { get { return this.storedFile; } }
        public string Description { get { return this.description; } }
        public ExtensionInfo Root { get { return this.root; } }
        public List<SlotInfo> RootSlots { get { return this.rootSlots; } }

        public SavedCompilationInfo(string filename, string description, ExtensionInfo root, List<SlotInfo> rootSlots)
        {
            this.storedFile = filename;
            this.description = description;
            this.root = root;
            this.rootSlots = rootSlots;
        }

        public int CompareTo(object obj)
        {
            SavedCompilationInfo sci = (SavedCompilationInfo)obj;
            return Description.CompareTo(sci.Description);
        }  
    }
  }
}